// MDDisplayManager
// by Steven M. Doyle <steve@macdweller.org>
//
// Canonical Home:  http://www.macdweller.org/library
//
// Intended to replace the undocumented O3Manager (at least partially) to allow
// for programmatic management of display brightness on both internal and
// external displays in a seamless fashion.
//
// Tested with the following:
//   - Mac Pro, 2006 - MacPro1,1 - OS 10.7.4 (11E53)
//      + Apple Cinema Display 30"
//
//   - 17" MacBook Pro, 2006 - MacBookPro2,1 - OS 10.7.3 (11D50)
//      + Internal Display
//
//   - Mac Pro, Early 2009 - MacPro4,1 - OS 10.8.1 (12B19)
//      + Apple Cinema Display 30"
//      + Apple Cinema Display 23"
//
//   - 20" iMac, Early 2009 - iMac9,1 - OS 10.7.4 (11E53)
//      + Internal Display
//
//   - 13" MacBook Air, Mid 2011 - MacBookAir4,2 - OS 10.8 (12A269)
//      + Internal Display
//      + Apple Cinema Display 30"
//
// Currently this seems to work pretty well, but any improvements or extensions
// are, of course, most welcome.  Please send patches c/o steve@macdweller.org.
//
// Feel free to use this code for whatever you want; it won't bother me so long
// as my name remains in the comments.  Of course, if you profit from it, I'm
// not so proud that I won't take donations. ;-)
//
// Release:  1.0

#import <Foundation/Foundation.h>


/// MDDisplayManager
///
/// Provides brightness controls and other minutae for connected displays, both
/// internal and external.  Displays are selected via their associated
/// CGDirectDisplayID value.  For example:
///
///   NSMutableArray *managers = [[NSMutableArray alloc] init];
///   for (NSScreen *display in [NSScreen screens]) {
///     CGDirectDisplayID dispid = [[[display deviceDescription]
///                                   objectForKey:@"NSScreenNumber"] intValue];
///     [managers addObject:[MDDisplayManager managerForDisplay:dispid]];
///   }
///
/// This will result in engines for all available displays.  Easy.  You can then
/// do various things; for example, to set the brightness of the first display
/// in the array:
///
///   MDDisplayManager *display = [managers objectAtIndex:0];
///   display.brightness = 0.25;
///
@interface MDDisplayManager : NSObject

/// Display brightness as a float between 0 and 1.0.
@property (assign, nonatomic) CGFloat brightness;

/// The display's product name according to the current locale.
@property (readonly) NSString *displayName;

/// The display's serial number (or 0 if unavailable, eg. an internal panel).
@property (readonly) unsigned int serialNumber;

/// Return the display manager for the specified display.
+ (id)managerForDisplay:(CGDirectDisplayID)disp;

/// Return the display's information dictionary.  This can be used to learn all
/// sorts of stuff.
- (NSDictionary *)infoDictionary;

@end
