// MDDisplayManager
// by Steven M. Doyle <steve@macdweller.org>
//
// Canonical Home:  http://www.macdweller.org/library
//
// Intended to replace the undocumented O3Manager (at least partially) to allow
// for programmatic management of display brightness on both internal and
// external displays in a seamless fashion.
//
// Tested with the following:
//   - Mac Pro, 2006 - MacPro1,1 - OS 10.7.4 (11E53)
//      + Apple Cinema Display 30"
//
//   - 17" MacBook Pro, 2006 - MacBookPro2,1 - OS 10.7.3 (11D50)
//      + Internal Display
//
//   - Mac Pro, Early 2009 - MacPro4,1 - OS 10.8.1 (12B19)
//      + Apple Cinema Display 30"
//      + Apple Cinema Display 23"
//
//   - 20" iMac, Early 2009 - iMac9,1 - OS 10.7.4 (11E53)
//      + Internal Display
//
//   - 13" MacBook Air, Mid 2011 - MacBookAir4,2 - OS 10.8 (12A269)
//      + Internal Display
//      + Apple Cinema Display 30"
//
// Currently this seems to work pretty well, but any improvements or extensions
// are, of course, most welcome.  Please send patches c/o steve@macdweller.org.
//
// Feel free to use this code for whatever you want; it won't bother me so long
// as my name remains in the comments.  Of course, if you profit from it, I'm
// not so proud that I won't take donations. ;-)
//
// Revision History
// ----------------
// 1.0 - Initial Release

/// @todo
///
/// * Test to see if HID Manager will update the list of matched devices without
///   being added to a runloop somewhere.  It's not on one at the moment, so I'm
///   not sure if it will respect device hot insertion/removal.
///
/// * Figure out the report parsing issue (or the obnoxious cache issue) in
///   MDDisplayManagerHIDImpl's setBrightness: method.
///
/// * Update managerForDisplay: such that we keep a single instance around for
///   any given display and simply return a reference to it.  This isn't as
///   simple as it sounds, however, because it needs to handle removal of the
///   cached object if its associated display is disconnected from the system.
///   I'm too lazy to deal with that at the moment.

#import <IOKit/graphics/IOGraphicsLib.h>
#import <IOKit/hid/IOHIDLib.h>
#import <mach/mach_time.h>
#import "MDDisplayManager.h"

#pragma mark - Constants & Macros

#define CFSTRDEF(x, y) static const CFStringRef x = CFSTR(y)
CFSTRDEF(kDisplayBrightness      , kIODisplayBrightnessKey);
CFSTRDEF(IOHIDPrimaryUsageKey    , kIOHIDPrimaryUsageKey);
CFSTRDEF(IOHIDPrimaryUsagePageKey, kIOHIDPrimaryUsagePageKey);
CFSTRDEF(IOHIDSerialNumberKey    , kIOHIDSerialNumberKey);
CFSTRDEF(IOHIDElementUsageKey    , kIOHIDElementUsageKey);
CFSTRDEF(IOHIDElementUsagePageKey, kIOHIDElementUsagePageKey);
CFSTRDEF(IOHIDElementReportIDKey , kIOHIDElementReportIDKey);
#undef CFSTRDEF

static const unsigned short HIDUSBMonitorUsagePage = 0x80;
static const unsigned short HIDMonitorControlUsage = 0x01;

#define NSSHORT(x) [NSNumber numberWithShort:x]


#pragma mark - Persistent Data


/// A reference to our global HID Manager instance.  Since the search parameters
/// are always the same, we'll use a single instance to handle all instances of
/// the USB HID implementation.
static IOHIDManagerRef hidManager = nil;


#pragma mark - MDIODisplayManager IODisplay Interface

/// MDDisplayManagerIODisplayImpl
///
/// Interface for the IODisplay-based implementation.
@interface MDDisplayManagerIODisplayImpl : MDDisplayManager {
    CGDirectDisplayID display;
    io_service_t      service;
}

@end


#pragma mark - MDHIDDisplayManager HID Interface

/// MDDisplayManagerHIDImpl
///
/// Interface for the USB HID-based implementation.
@interface MDDisplayManagerHIDImpl : MDDisplayManagerIODisplayImpl {
    IOHIDDeviceRef  hidDevice;
    IOHIDElementRef hidBrightnessElement; 
}
@end


#pragma mark - MDDispayManager Internal Method Category


/// MDDisplayManager (Internal)
///
/// Category providing internal method declarations for stuff we don't want
/// exposed in the header.
@interface MDDisplayManager (Internal)
- initWithDisplay:(CGDirectDisplayID)disp service:(io_service_t)svc;
@end


#pragma mark - MDDisplayManager Base Implementation

/// MDDisplayManager
///
/// Base class for all implementations; the only functionality it provides is
/// in selecting the proper class to implement controls for the specified
/// display.
@implementation MDDisplayManager

/// managerForDisplay:
///
/// Creates the appropriate MDDisplayManager for the display associated with the
/// specified CGDirectDisplayID.
///
/// Wouldn't it be nice if Objective-C supported true abstract classes?
+ (id)managerForDisplay:(CGDirectDisplayID)disp {
    // Check to see if we can gather information via IODisplay method.  If so,
    // hand back and implementation using that method.
    io_service_t service = CGDisplayIOServicePort(disp);
    float brightness;
    if (kIOReturnSuccess == IODisplayGetFloatParameter(service,
                                                       kNilOptions,
                                                       kDisplayBrightness,
                                                       &brightness))
        return [[[MDDisplayManagerIODisplayImpl alloc] initWithDisplay:disp
                                                               service:service] autorelease];
    
    
    // If we're still here, then the only other option is HID.
    return [[[MDDisplayManagerHIDImpl alloc] initWithDisplay:disp
                                                     service:service] autorelease];
}


- (id)initWithDisplay:(CGDirectDisplayID)disp service:(io_service_t)svc { return [super init]; }
- (CGFloat)brightness { return 0; }
- (void)setBrightness:(CGFloat)brightness { }
- (NSString *)displayName { return @"Unsupported"; }
- (unsigned int)serialNumber { return 0; }
- (NSDictionary *)infoDictionary { return [NSDictionary dictionary]; }

@end


#pragma mark - MDDisplayManager IODisplay Implementation

/// MDDisplayManagerIODisplayImpl
///
/// This implementation relies on the IODisplay subsystem to talk to the display
/// controls.  It generally only works with internal displays (eg. on laptops).
/// It also serves as a base class for the USB HID implementation, which still
/// requires some IODisplay functionality.
///
/// This code has been tested against internal displays on the following Mac
/// models:
///   - MacBook Air 13" (Mid 2011, MacBookAir4,2)
///   - MacBook Pro 17" (2006?)
///
@implementation MDDisplayManagerIODisplayImpl


/// initWithDisplay:service:
///
/// Initialize the HID driver.
- (id)initWithDisplay:(CGDirectDisplayID)disp service:(io_service_t)svc {
    if (self = [super init]) {
        display = disp;
        service = svc;
    }
    return self;
}


/// displayName
///
/// Return the product name associated with this display (eg. "Cinema HD") as
/// appropriate for the current locale.
- (NSString *)displayName {
    NSDictionary *devInfo = [(NSDictionary *)IODisplayCreateInfoDictionary(service, kIODisplayOnlyPreferredName) autorelease];    
    NSDictionary *localNames = [devInfo objectForKey:[NSString stringWithUTF8String:kDisplayProductName]];
    return [[[localNames objectForKey:[[NSLocale currentLocale] localeIdentifier]] retain] autorelease];
}


/// infoDictionary
///
/// Return an autoreleased copy of the Display's information dictionary as
/// provided by the IODisplay subsystem.
- (NSDictionary *)infoDictionary {
    return [(NSDictionary *)IODisplayCreateInfoDictionary(service, 0) autorelease];
}


/// serialNumber
///
/// Return the serial number associated with this display as an unsigned
/// integer.  Some displays (such as the internal display on Apple laptops) will
/// return 0.
- (unsigned int)serialNumber {
    return [[[self infoDictionary] objectForKey:@"DisplaySerialNumber"] unsignedIntValue];
}


/// brightness
///
/// Retrieve the current brightness value for the associated display.  Returns
/// a float between 0 (dark) and 1.0 (bright).  If the brightness can't be
/// retreived for some reason, it will return 0.
- (CGFloat)brightness {
    CGFloat brightness;
    
    if (kIOReturnSuccess != IODisplayGetFloatParameter(service,
                                                       kNilOptions,
                                                       kDisplayBrightness,
                                                       &brightness)) {
        NSLog(@"Failed to get brightness value [display=%d]", display);
        return 0;
    }
    
    return brightness;
}


/// setBrightness:
///
/// Sets the brightness level for the associated display.  The input value must
/// be a floating point value between 0 (dark) and 1.0 (bright).
- (void)setBrightness:(CGFloat)brightness {
    if (kIOReturnSuccess != IODisplaySetFloatParameter(service,
                                                       kNilOptions,
                                                       kDisplayBrightness,
                                                       brightness))
        NSLog(@"Failed to set display brightness:  [display=%d]", display);
}

@end


#pragma mark - MDDisplayManager HID Implementation
/// MDDisplayManagerHIDImpl
///
/// This implementation relies on the USB HID subsystem to talk to the display
/// controls.  This should work with any display that supports the VESA Monitor
/// control usage pages, but I make no promises.
///
/// We're derived from the IODisplay implementation because it gives us a fully
/// consistent method for grabbing the product name.  If functionality ever
/// expands, I expect that the IODisplay code will play a very large part as
/// well.
///
/// This code has been tested with the following displays:
///   - Apple 30" Cinema Display
///   - Apple 23" Cinema Display
///
@implementation MDDisplayManagerHIDImpl


/// initWithDisplay:service:
///
/// Initialize the HID driver.
- (id)initWithDisplay:(CGDirectDisplayID)disp service:(io_service_t)svc {
    /// @todo
    ///
    /// The usagePage/usage pair constants are missing here because I need to
    /// back through the USB HID documentation and figure out WTF everything
    /// should be called.
    ///
    /// Been a while since I looked into HID report descriptors...
    if (self = [super initWithDisplay:disp service:svc]) {
        if ((hidDevice = [self hidSetup])) {
            CFRetain(hidDevice);

            
            hidBrightnessElement = [self hidCopyElementForDevice:hidDevice
                                                       usagePage:0x82
                                                           usage:0x10];
        }
    }
    return self;
}


/// dealloc
///
/// This code is not ARC compliant, and probably won't be any time soon, so we
/// still need this to clean up after ourselves.
- (void)dealloc {
    if (hidBrightnessElement)
        CFRelease(hidBrightnessElement);
    
    if (hidDevice) {
        CFRelease(hidDevice);
    }
    
    [super dealloc];
}


/// hidManagerSetup
///
/// Create our global HID Manager object and initialize it to match with devices
/// that support USB monitor control functionality.
- (void)hidManagerSetup {
    static dispatch_once_t hms_pred;
    dispatch_once(&hms_pred, ^{        
        // Create the HID Manager.
        hidManager = IOHIDManagerCreate(kCFAllocatorDefault, kIOHIDOptionsTypeNone);
    
        // Set device matching for only devices with a primary USB Monitor/
        // Monitor Control usage pair.
        NSDictionary *match = [NSDictionary dictionaryWithObjectsAndKeys:
                               NSSHORT(HIDMonitorControlUsage), IOHIDPrimaryUsageKey,
                               NSSHORT(HIDUSBMonitorUsagePage), IOHIDPrimaryUsagePageKey,
                               nil];
        IOHIDManagerSetDeviceMatching(hidManager, (CFDictionaryRef)match);

        // Open the HID Manager.
        if (kIOReturnSuccess != IOHIDManagerOpen(hidManager, kIOHIDOptionsTypeNone))
            NSLog(@"WARNING:  Failed to open HID manager.");
    });
}


/// hidSetup
///
/// Initialize the USB HID subsystem and locate the device for the monitor we
/// currently represent.
///
/// This routine takes the monitor serial number associated with our assigned
/// CGDirectDisplayID, finds a HID device of appropriately functionality that
/// sports the same serial number, and returns it.  For this to work properly,
/// the serial numbers must match; this seems to work on Apple-branded displays,
/// but there's no guarantee on third party units.
///
/// Returns the IOHIDDeviceRef of the matching HID device, or nil if it can't
/// find one.
- (IOHIDDeviceRef)hidSetup {
    // Make sure the hid manager is initialized.
    [self hidManagerSetup];
    
    
    // Take a gander through the devices and see if we can find one that matches
    // the serial number of the display we're looking for.  Note that the USB
    // HID serial number key is a string key; we therefore have to interpret it
    // appropriately to compare against our integer serial number (sigh).
    NSSet *devices = (NSSet *)IOHIDManagerCopyDevices(hidManager);
    [devices autorelease];
    
    for (id item in devices) {
        NSString *snstr = IOHIDDeviceGetProperty((IOHIDDeviceRef)item,
                                                 IOHIDSerialNumberKey);

        // Convert the hexadecimal serial number to an unsigned int
        NSScanner *scan = [NSScanner scannerWithString:snstr];
        unsigned int test;
        [scan scanHexInt:&test];
        
        // See if we match, and if so, return the device.
        if (test == self.serialNumber)
            return (IOHIDDeviceRef)item;
    }
    
    
    // No device found.
    return nil;
}


/// hidCopyElementForDevice:usagePage:usage:
///
/// Locates and returns the HID element associated with the specified usage
/// pair for the specified device.
- (IOHIDElementRef)hidCopyElementForDevice:(IOHIDDeviceRef)device
                                 usagePage:(unsigned short)page
                                     usage:(unsigned short)usage {
    NSDictionary *match = [NSDictionary dictionaryWithObjectsAndKeys:
                           NSSHORT(page), IOHIDElementUsagePageKey,
                           NSSHORT(usage), IOHIDElementUsageKey,
                           nil];
    
    NSArray *elements = (NSArray *)IOHIDDeviceCopyMatchingElements(device,
                                                                   (CFDictionaryRef)match,
                                                                   kIOHIDOptionsTypeNone);
    
    if (!elements.count) {
        NSLog(@"Warning:  brightness element not found.");
        CFRelease(elements);
        return nil;
    }
    
    if (elements.count > 1)
        NSLog(@"Warning:  more than one brightness element located.");
    
    // Got an element; retain the one we want and ditch the rest.
    IOHIDElementRef result = (IOHIDElementRef)[elements objectAtIndex:0];
    CFRetain(result);
    CFRelease(elements);
    return result;
}


/// brightness
///
/// Retrieve the current brightness value for the associated display.  Returns
/// a float between 0 (dark) and 1.0 (bright).  If the brightness can't be
/// retreived for some reason, it will return 0.
- (CGFloat)brightness {
    if (!hidDevice || !hidBrightnessElement)
        return 0;
    
    /// @todo
    ///
    /// IOHIDDeviceGetValue() does not always actually *send* the request for
    /// the report we're interested in, and instead uses a cached copy.  This
    /// results in abject failure; while changes at the monitor will be
    /// reflected, changes via IOHIDDeviceSetValue() are *not* for some reason.
    ///
    /// Yeah, it's a thing.
    ///
    /// The only way I found to work around this is to manually request the
    /// report instead of using the higher-level API calls.  In the monitors I
    /// have, the brightness element is the sole member of its report, so it
    /// works just fine.
    ///
    /// YMMV on other monitors.
    ///
    /// At some point in the future, we should either figure out how to
    /// invalidate the internal cache in the IOHID stuff, or build a proper
    /// report handler that can extract the data according to the elements in
    /// the HID descriptor.  It would make things much less likely to break.
    ///
    /// Of course, I'm way too lazy to do that at the moment, so behold the
    /// hardcoded report structure.  The only consession is some extra space in
    /// case the overall report is larger than we expect, which should help to
    /// prevent buffer overruns (maximum packet size for HID is 64 bytes if
    /// memory serves; if it sends a multi-packet report, you're on your own).
    CFIndex len = IOHIDElementGetReportSize(hidBrightnessElement);
    struct {
        uint8_t report_id;
        uint8_t brightness;
        uint8_t garbage[62];
    } report;
    
    if (kIOReturnSuccess != IOHIDDeviceGetReport(hidDevice,
                                                 kIOHIDReportTypeFeature,
                                                 IOHIDElementGetReportID(hidBrightnessElement),
                                                 (uint8_t *)&report,
                                                 &len)) {
        NSLog(@"Failed to get brightness value [display=%d]", display);
        return 0;
    }
    
    return report.brightness / 255.0;
    
#if 0
    /// This is the original code.  It reads the value the first time after the
    /// HID Manager is opened, and then repeats that value until the report
    /// containing the brightness value is sent back by the monitor due to some
    /// event (like the hard buttons on the monitor).  No matter how many times
    /// you do it, IOHIDSetDeviceValue() will not invalidate the cached report,
    /// which is (frankly) stupid.
    IOHIDValueRef value;
    if (kIOReturnSuccess != IOHIDDeviceGetValue(hidDevice,
                                                hidBrightnessElement, &value)) {
        NSLog(@"Failed to get brightness value [display=%d]", display);
        return 0;
    }
        
    return IOHIDValueGetIntegerValue(value) / 255.0;
#endif
}


/// setBrightness:
///
/// Sets the brightness level for the associated display.  The input value must
/// be a floating point value between 0 (dark) and 1.0 (bright).
- (void)setBrightness:(CGFloat)brightness {
    if (!hidDevice || !hidBrightnessElement)
        return;

    IOHIDValueRef value = IOHIDValueCreateWithIntegerValue(CFAllocatorGetDefault(),
                                                           hidBrightnessElement,
                                                           mach_absolute_time(),
                                                           brightness * 255);
    
    if (kIOReturnSuccess != IOHIDDeviceSetValue(hidDevice, hidBrightnessElement, value))
        NSLog(@"Failed to set display brightness:  [display=%d]", display);

    CFRelease(value);
}


@end